using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.Text;

namespace Serialization
{
    [DataContract(Name="Calculation")]
    public class ServiceViewOfData: IUnknownSerializationData
    {
        [DataMember(Name = "Symbols")]
        private string[] symbols;
        [DataMember(Name = "Parameters")]
        private decimal[] parameters;
        [DataMember(Name = "Functions")]
        private string[] functions;
        [DataMember(Name="Value")]
        private decimal value;
        
        private UnknownSerializationData unknownData;

        public string[] Symbols
        {
            get
            {
                return this.symbols;
            }

            set
            {
                this.symbols = value;
            }
        }

        public decimal[] Parameters
        {
            get
            {
                return this.parameters;
            }

            set
            {
                this.parameters = value;
            }
        }

        public string[] Functions
        {
            get
            {
                return this.functions;
            }

            set
            {
                this.functions = value;
            }

        }

        public decimal Value
        {
            get
            {
                return this.value;
            }

            set
            {
                this.value = value;
            }
        }

        #region IUnknownSerializationData Members

        public UnknownSerializationData UnknownData
        {
            get
            {
                return this.unknownData;
            }

            set
            {
                this.unknownData = value;
            }
        }

        #endregion
    }

    [DataContract]
    public class Data 
    {
        [DataMember]
        public string Value;
    }

    [DataContract]
    public class DerivedData : Data
    {
    }

    [DataContract(Name = "Calculation")]
    public class ClientViewOfData : IUnknownSerializationData
    {
        [DataMember(Name = "Symbols")]
        public string[] Symbols;
        [DataMember(Name = "Parameters")]
        public decimal[] Parameters;
        [DataMember(Name="Functions")]
        public string[] Functions;
        [DataMember(Name="Value")]
        public decimal Value;
        [DataMember(Name = "Reference")]
        public Guid Reference;

        private UnknownSerializationData unknownData;


        public UnknownSerializationData UnknownData
        {
            get
            {
                return this.unknownData;
            }

            set
            {
                this.unknownData = value;
            }
        }


    }


    

    [ServiceContract(Name = "DerivativesCalculator")]
    [KnownType(typeof(DerivedData))]
    public interface IServiceViewOfService
    {
        [OperationContract(Name="FromXSDTypes")]
        decimal CalculateDerivative(
            string[] symbols,
            decimal[] parameters,
            string[] functions);

        [OperationContract(Name="FromDataSet")]
        DataSet CalculateDerivative(DataSet input);

        [OperationContract(Name = "FromDataContract")]
        ServiceViewOfData CalculateDerivative(ServiceViewOfData input);

        [OperationContract(Name = "AlsoFromDataContract")]
        Data DoSomething(Data input);
        
    }

    [ServiceContract(Name="DerivativesCalculator")]
    [KnownType(typeof(DerivedData))]
    public interface IClientViewOfService
    {
        [OperationContract(Name = "FromXSDTypes")]
        decimal CalculateDerivative(
            string[] symbols,
            decimal[] parameters,
            string[] functions);

        [OperationContract(Name = "FromDataSet")]
        DataSet CalculateDerivative(DataSet input);

        [OperationContract(Name = "FromDataContract")]
        ClientViewOfData CalculateDerivative(ClientViewOfData input);

        [OperationContract(Name = "AlsoFromDataContract")]
        Data DoSomething(Data input);
    }

    public class DerivativesCalculator : IServiceViewOfService
    {
        #region IDerivativesCalculator Members

        public decimal CalculateDerivative(
            string[] symbols, 
            decimal[] parameters, 
            string[] functions)
        {
            return (decimal)(System.DateTime.Now.Millisecond);
        }
    

        public DataSet CalculateDerivative(DataSet input)
        {
            if (input.Tables.Count > 0)
            {
                if (input.Tables[0].Rows.Count > 0)
                {
                    input.Tables[0].Rows[0]["Value"] = 
                        (decimal)(System.DateTime.Now.Millisecond);
                }
            }
            return input;
        }

        public ServiceViewOfData CalculateDerivative(ServiceViewOfData input)
        {
            input.Value = this.CalculateDerivative(
                input.Symbols, 
                input.Parameters, 
                input.Functions);
            return input;
        }

        public Data DoSomething(Data input)
        {
            return input;
        }

        #endregion
    }

    public class Program
    {
        public static void Main(string[] args)
        {
            using (ServiceHost host = new ServiceHost(
                typeof(DerivativesCalculator), 
                new Uri[] { 
                    new Uri("http://localhost:8000/Derivatives") }))
            {
                host.AddServiceEndpoint(
                    typeof(IServiceViewOfService), 
                    new BasicHttpBinding(), 
                    "Calculator");
                host.Open();

                Console.WriteLine("The service is available.");

                string address =
                    "http://localhost:8000/Derivatives/Calculator";
                ChannelFactory<IClientViewOfService> factory =
                    new ChannelFactory<IClientViewOfService>(
                        new BasicHttpBinding(),
                        new EndpointAddress(
                            new Uri(address)));
                IClientViewOfService proxy =
                    factory.CreateChannel();

                decimal result = proxy.CalculateDerivative(
                    new string[] { "MSFT" },
                    new decimal[] { 3 },
                    new string[] { "TechStockProjection" });
                Console.WriteLine("Value using XSD types is {0}.", result);

                DataTable table = new DataTable("InputTable");
                table.Columns.Add("Symbol", typeof(string));
                table.Columns.Add("Parameter", typeof(decimal));
                table.Columns.Add("Function", typeof(string));
                table.Columns.Add("Value", typeof(decimal));

                table.Rows.Add("MSFT", 3, "TechStockProjection",0.00);
                
                DataSet input = new DataSet("Input");
                input.Tables.Add(table);

                DataSet output = proxy.CalculateDerivative(input);
                if (output != null)
                {
                    if (output.Tables.Count > 0)
                    {
                        if (output.Tables[0].Rows.Count > 0)
                        {
                            Console.WriteLine(
                                "Value using a DataSet is {0}.",
                                output.Tables[0].Rows[0]["Value"]);
                        }
                    }
                }

                ClientViewOfData calculation = 
                    new ClientViewOfData();
                calculation.Symbols = 
                    new string[] { "MSFT" };
                calculation.Parameters = 
                    new decimal[] { 3 };
                calculation.Functions = 
                    new string[] { "TechStockProjection" };
                calculation.Reference =
                    Guid.NewGuid();

                Console.WriteLine(
                    "Reference is {0}.", calculation.Reference);

                ClientViewOfData calculationResult = 
                    proxy.CalculateDerivative(calculation);
                Console.WriteLine("Value using a Data Contract {0}.", 
                    calculationResult.Value);

                Console.WriteLine(
                    "Reference is {0}.", calculationResult.Reference);

                DerivedData derivedData = new DerivedData();
                Data outputData = proxy.DoSomething(derivedData);

                MemoryStream stream = new MemoryStream();
                XmlFormatter formatter = new XmlFormatter();
                formatter.Serialize(stream, calculation);
                Console.WriteLine(
                    UnicodeEncoding.UTF8.GetChars(stream.GetBuffer()));

                Console.WriteLine("Done.");

                ((IChannel)proxy).Close();

                host.Close();

                Console.ReadKey();

            }
        }
    }
}
