using System;
using System.Configuration;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Collections.Generic;
using System.Text;
using System.ServiceModel;
using System.ServiceModel.Security;
using System.ServiceModel.Security.Protocols;
using System.ServiceModel.Security.Tokens;
using System.Security.Authorization;

namespace SecurityTokenService
{
    class DoNothingX509Validator : X509CertificateValidator
    {
        static DoNothingX509Validator instance = new DoNothingX509Validator();

        public static DoNothingX509Validator Instance
        {
            get { return instance; }
        }

        public override void Validate(X509Certificate2 certificate)
        {
        }
    }

    public class SecurityTokenService : SecurityTokenServiceBase
    {
        X509Certificate2 signatureCertificate;
        X509AsymmetricSecurityKey signatureKey;
        SecurityKeyIdentifier signatureKeyIdentifier;
        string issuerName;

        public SecurityTokenService():base(true)
        {
            string issuerName = ConfigurationManager.AppSettings["IssuerName"];
            X509Store store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
            store.Open(OpenFlags.ReadOnly);
            X509Certificate2Collection certs = store.Certificates.Find(X509FindType.FindBySubjectName, issuerName, false);
            if (certs.Count != 1)
                throw new Exception("Unable to find exception for '" + issuerName + "'");
            this.issuerName = issuerName;
            this.signatureCertificate = certs[0];
            this.signatureKey = new X509AsymmetricSecurityKey(signatureCertificate);
            this.signatureKeyIdentifier = new SecurityKeyIdentifier(new X509ThumbprintKeyIdentifierClause(this.signatureCertificate));
        }

        byte[] GenerateProofKey(RequestSecurityToken request, RequestSecurityTokenResponse response)
        {
            int keySize;
            if (request.KeySize != 0)
                keySize = request.KeySize;
            else
                keySize = 256;
            response.Context = request.Context;
            response.KeySize = keySize;
            response.TokenType = request.TokenType;
            byte[] senderEntropy = null;
            SecurityToken entropyToken = request.GetRequestorEntropy();
            if (entropyToken != null)
            {
                senderEntropy = ((BinarySecretSecurityToken)entropyToken).GetKeyBytes();
            }
            byte[] key;
            if (senderEntropy != null)
            {
                byte[] stsEntropy = new byte[keySize / 8];
                RNGCryptoServiceProvider random = new RNGCryptoServiceProvider();
                random.GetNonZeroBytes(stsEntropy);
                key = RequestSecurityTokenResponse.ComputeCombinedKey(senderEntropy, stsEntropy, keySize);
                response.SetIssuerEntropy(stsEntropy);
                response.ComputeKey = true;
            }
            else
            {
                key = new byte[keySize / 8];
                RNGCryptoServiceProvider random = new RNGCryptoServiceProvider();
                random.GetNonZeroBytes(key);
            }
            response.RequestedProofToken = new BinarySecretSecurityToken(key);
            return key;
        }

        public override RequestSecurityTokenResponse Issue(RequestSecurityToken request)
        {
            RequestSecurityTokenResponse response = new RequestSecurityTokenResponse(SecurityStandardsManager.DefaultInstance);
            byte[] proofKey = GenerateProofKey(request, response);

            // TODO: This should come from the request...
            X509SecurityToken wrappingKey = new X509SecurityToken(signatureCertificate, DoNothingX509Validator.Instance, SecurityToken.GenerateId());
            wrappingKey.Validate();
            SecurityKeyIdentifier proofKeyIdentifier = WrapProofKey(request, proofKey);

            DateTime effectiveTime = DateTime.Now;
            DateTime expirationTime = effectiveTime.AddDays(1);
            response.SetLifetime(effectiveTime, expirationTime);

            IList<Claim> claims = new List<Claim>();

            foreach (ClaimSet cs in OperationContext.Current.ServiceSecurityContext.AuthorizationContext.ClaimSets)
                foreach (Claim c in cs)
                    claims.Add(c);

            ClaimSet claimSet = new DefaultClaimSet(ClaimSet.System, claims);

            response.RequestedSecurityToken = CreateSamlToken(
                SecurityAlgorithms.RsaSha1Signature,
                SecurityAlgorithms.Sha1Digest,
                effectiveTime,
                expirationTime,
                claimSet,
                proofKeyIdentifier);

            return response;
        }

        SecurityKeyIdentifier WrapProofKey(RequestSecurityToken request, byte[] proofKey)
        {
            EndpointAddress address = request.GetAppliesTo<EndpointAddress>();
            string identityClaim = address.Identity.IdentityClaim.Resource.ToString();
            X509Store store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
            store.Open(OpenFlags.ReadOnly);
            X509Certificate2Collection certs = store.Certificates.Find(X509FindType.FindBySubjectName, identityClaim, false);
            if (certs.Count != 1)
            {
                throw new Exception("Cert not found for " + address.Uri.Host);
            }
            X509SecurityToken wrappingKey = new X509SecurityToken(certs[0], DoNothingX509Validator.Instance, SecurityToken.GenerateId());
            wrappingKey.Validate();
            SecurityKeyIdentifier wrappingKeyIdentifier = new SecurityKeyIdentifier(wrappingKey.CreateKeyIdentifierClause<X509ThumbprintKeyIdentifierClause>());
            byte[] wrappedKey = wrappingKey.SecurityKeys[0].EncryptKey(SecurityAlgorithmSuite.Default.DefaultAsymmetricKeyWrapAlgorithm, proofKey);
            return new SecurityKeyIdentifier(new EncryptedKeyIdentifierClause(wrappedKey, SecurityAlgorithmSuite.Default.DefaultAsymmetricKeyWrapAlgorithm, wrappingKeyIdentifier));
        }

        SamlSecurityToken CreateSamlToken(
            string signatureAlgorithm,
            string digestAlgorithm,
            DateTime effectiveTime,
            DateTime expirationTime,
            ClaimSet claimSet,
            SecurityKeyIdentifier proofKeyIdentifier)
        {
            SamlSubject subject = new SamlSubject(
                null,
                null,
                "Some User",
                new string[] { "UsernamePassword" },
                null,
                proofKeyIdentifier);

            List<SamlAttribute> attributes = new List<SamlAttribute>();

            if (claimSet == null)
                claimSet = new DefaultClaimSet(ClaimSet.System, Claim.CreateRoleClaim("Anonymous"));

            for (int i = 0; i < claimSet.Count; i++)
                if (claimSet[i].Right == Rights.PossessProperty)
                    attributes.Add(new SamlAttribute(claimSet[i].ClaimType.Substring(0, claimSet[i].ClaimType.LastIndexOf('/') + 1), claimSet[i].ClaimType.Substring(claimSet[i].ClaimType.LastIndexOf('/') + 1), new string[] { claimSet[i].Resource.ToString() }));

            SamlAttributeStatement attributeStatement = new SamlAttributeStatement(
                subject,
                attributes);

            SamlAssertion assertion = new SamlAssertion(
                SecurityToken.GenerateId(),
                issuerName,
                DateTime.Now,
                new SamlConditions(DateTime.Now, DateTime.Now.AddDays(1)),
                null,
                new SamlStatement[] { attributeStatement });

            assertion.SigningCredentials = new SigningCredentials(
                signatureKey,
                signatureAlgorithm,
                digestAlgorithm,
                signatureKeyIdentifier);

            SamlSecurityToken token = new SamlSecurityToken(assertion);

            return token;
        }
    }

    
}

