using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;

namespace PublicationSubscription
{
	public partial class NotificationStreamWriter: Stream, INotificationStreamWriter
	{
		private IBufferedSubscriptionManager bufferedDataSubscriptionManager = null;
		private long messageCapacity;
		private TimeSpan sourceTimeout;
		private TimeSpan updateFrequency;
		private bool initialRead = true;
		private long totalBytesRead = 0;
		private ManualResetEvent waitHandle = null;
		private Timer updateTimer = null;
		private bool updateRequired = false;

		private object waitHandleLock = null;
		private object timerLock = null;
		private object updateRequiredLock = null;

		private NotificationStreamWriter()
		{
		}

		public NotificationStreamWriter(IBufferedSubscriptionManager bufferedDataSubscriptionManager, long messageCapacity, TimeSpan updateFrequency, TimeSpan sourceTimeout)
		{
			this.bufferedDataSubscriptionManager = bufferedDataSubscriptionManager;
			this.messageCapacity = messageCapacity;
			this.updateFrequency = updateFrequency;
			this.sourceTimeout = sourceTimeout;

			this.bufferedDataSubscriptionManager.DataAvailable += new DataAvailableHandler(this.DataAvailableHandler);

			this.waitHandle = new ManualResetEvent(false);

			this.waitHandleLock = new object();
			this.timerLock = new object();
			this.updateRequiredLock = new object();
		}

		private void DataAvailableHandler(object sender)
		{
			lock (waitHandleLock)
			{
				this.waitHandle.Set();
			}
		}

		private void TimerHandler(object state)
		{
			lock (this.updateRequiredLock)
			{
				this.updateRequired = true;
			}

			this.DisposeOfUpdateTimer();
		}

		private void DisposeOfUpdateTimer()
		{
			lock (this.timerLock)
			{
				if (this.updateTimer != null)
				{
					this.updateTimer.Dispose();
					this.updateTimer = null;
				}

			}
		}

		private int ReadFromBuffer(byte[] sourceBuffer, byte[] outputBuffer, int offset)
		{
			Array.Copy(sourceBuffer, 0, outputBuffer, offset, sourceBuffer.Length);

			return sourceBuffer.Length;
		}

		private int InitialRead(byte[] buffer, int offset, int count)
		{
			byte initialReadSize = 1;

			if (buffer.Length < initialReadSize)
			{
				throw new ArgumentException();
			}

			buffer[0] = SubscriptionStreamControlSignal.Size;

			this.initialRead = false;

			return initialReadSize;
		}

		private int SubsequentRead(byte[] buffer, int offset, int count)
		{
			int bytesRead = 0;
			IBufferedDataPointValue value = null;
			ISubscriptionStreamControlSignal controlSignal = null;
			int offsetAfterNextValue = 0;
			int sizeOfNextRead = 0;

			lock (this.updateRequiredLock)
			{
				this.updateRequired = false;
			}

			lock (this.timerLock)
			{
				this.updateTimer = new Timer(this.TimerHandler, null, this.updateFrequency, new TimeSpan(0, 0, 0, 0, -1));
			}

			while (true)
			{
				if (this.bufferedDataSubscriptionManager.DataPointValueBuffer.Count > 0)
				{
					value = this.bufferedDataSubscriptionManager.DataPointValueBuffer.Peek();
					sizeOfNextRead = SubscriptionStreamControlSignal.Size + value.Value.Length;
					offsetAfterNextValue = offset + bytesRead + sizeOfNextRead;
					if ((offsetAfterNextValue > buffer.Length)||(offsetAfterNextValue > count)||((this.totalBytesRead + bytesRead + sizeOfNextRead) > this.messageCapacity))
					{
						Console.WriteLine("Reached a limit.");
						break;
					}

					value = this.bufferedDataSubscriptionManager.DataPointValueBuffer.Dequeue();

					controlSignal = new SubscriptionStreamControlSignal(value.DataPointIdentifier, value.Value.Length);
					bytesRead += this.ReadFromBuffer(controlSignal.ToByteArray(), buffer, offset + bytesRead);

					bytesRead += this.ReadFromBuffer(value.Value, buffer, offset + bytesRead);

					Console.WriteLine("Wrote value {0} of data point {1}.", BitConverter.ToInt32(value.Value,0), value.DataPointIdentifier);

				}
				else
				{
					lock (this.waitHandleLock)
					{
						this.waitHandle.Reset();
					}
					this.waitHandle.WaitOne();

					if (!(this.bufferedDataSubscriptionManager.DataPointValueBuffer.Count > 0))
					{
						break;
					}
				}

				lock (this.updateRequiredLock)
				{
					if (this.updateRequired)
					{
						break;
					}
				}
			}

			this.DisposeOfUpdateTimer();

			return bytesRead;
		}
				
		public override int Read(byte[] buffer, int offset, int count)
		{
			Console.WriteLine("Reading.");

			int bytesRead = 0;

			if (buffer.Length < offset + count)
			{
				throw new ArgumentException();
			}
						
			if (this.initialRead)
			{
				bytesRead += this.InitialRead(buffer, offset, count);
			}
			else
			{
				bytesRead += this.SubsequentRead(buffer, offset, count);
			}
			
			this.totalBytesRead += bytesRead;

			Console.WriteLine("Bytes read so far: {0}", totalBytesRead);

			return bytesRead;
		}

		#region INotificationStream Members

		long INotificationStreamWriter.MessageCapacity
		{
			get 
			{
				return this.messageCapacity;
			}
		}

		IBufferedSubscriptionManager INotificationStreamWriter.BufferedDataSubscriptionManager
		{
			get 
			{
				return this.bufferedDataSubscriptionManager;
			}
		}

		
		TimeSpan INotificationStreamWriter.Timeout
		{
			get 
			{
				return this.sourceTimeout;
			}
		}

		TimeSpan INotificationStreamWriter.UpdateFrequency
		{
			get 
			{ 
				return this.updateFrequency;	
			}
		}

		#endregion
}
}
