using System.IO;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Design;
using System.ServiceModel.Security.Protocols;
using System.Text;
using System.Xml;

namespace System.ServiceModel
{
	public class Constants
	{
		public const string DefaultAction = "*";
		public const string HttpRequestMessageProperty = "HttpeRequestMessageProperty";
	}

	public class PlainXMLEncoderBindingElement : BindingElement, IMessageEncodingBindingElement
	{
        public override IChannelFactory BuildChannelFactory(ChannelBuildContext context)
        {
            if (context == null)
                throw new ArgumentNullException("context");

            context.UnhandledBindingElements.Add(this);
            return context.BuildInnerChannelFactory();
        }

        public override IChannelListener<TChannel> BuildChannelListener<TChannel>(ChannelBuildContext context)
        {
            if (context == null)
                throw new ArgumentNullException("context");

            context.UnhandledBindingElements.Add(this);
            return context.BuildInnerChannelListener<TChannel>();
        }

        public override bool CanBuildChannelListener<TChannel>(ChannelBuildContext context)
        {
            if (context == null)
                throw new ArgumentNullException("context");

            context.UnhandledBindingElements.Add(this);
            return context.CanBuildInnerChannelListener<TChannel>();
        }

        public override BindingElement Clone()
		{
			return new PlainXMLEncoderBindingElement();
		}

		public override ChannelProtectionRequirements GetProtectionRequirements()
		{
			throw new Exception("The method or operation is not implemented.");
		}

		#region IMessageEncodingBindingElement Members

		AddressingVersion IMessageEncodingBindingElement.AddressingVersion
		{
			get { throw new Exception("The method or operation is not implemented."); }
		}

		MessageEncoderFactory IMessageEncodingBindingElement.CreateMessageEncoderFactory()
		{
			return new PlainXMLMessageEncoderFactory();
		}

		#endregion
	}

	public class PlainXMLMessageEncoderFactory : MessageEncoderFactory
	{

		public override MessageEncoder Encoder
		{
			get
			{
				return new PlainXMLEncoder();
			}
		}

		public override MessageVersion MessageVersion
		{
			get
			{
				return MessageVersion.Soap11Addressing1;
			}
		}
	}


	public class PlainXMLEncoder : MessageEncoder
	{
		private const string xmlTextMediaType = "text/xml";
		private const string xmlTextContentType = "text/xml; charset=utf-8";


		public override string MediaType
		{
			get
			{
				return PlainXMLEncoder.xmlTextMediaType;
			}
		}

		public override string ContentType
		{
			get
			{
				return PlainXMLEncoder.xmlTextContentType;
			}
		}

		public override MessageVersion MessageVersion
		{
			get
			{
				return MessageVersion.Soap11Addressing1;
			}

		}

		public override Message ReadMessage(ArraySegment<byte> buffer, BufferManager bufferManager)
		{
			Message message = null;

			if (buffer.Count > 0)
			{
				MemoryStream stream = new MemoryStream(buffer.Array);
				XmlTextReader reader = new XmlTextReader(stream);
				message = new PlainXMLMessage(MessageVersion.Soap11Addressing1, Constants.DefaultAction, reader);
				return message;
			}

			message = new PlainXMLMessage(MessageVersion.Soap11Addressing1, Constants.DefaultAction);

			return message;
		}

		public override ArraySegment<byte> WriteMessage(Message message, int maxMessageSize, BufferManager bufferManager, int messageOffset)
		{
			byte[] buffer = bufferManager.TakeBuffer(maxMessageSize);
			MemoryStream stream = new MemoryStream(buffer);
			XmlTextWriter innerWriter = new XmlTextWriter(stream, Encoding.UTF8);
			XmlDictionaryWriter outerWriter = XmlDictionaryWriter.CreateDictionaryWriter(innerWriter, false);
			message.WriteBodyContents(outerWriter);
			outerWriter.Flush();
			int position = (int)stream.Position;
			outerWriter.Close();
			return new ArraySegment<byte>(buffer, 0, position);

		}

		#region stream functions

		public override void WriteMessage(Message message, System.IO.Stream stream)
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override Message ReadMessage(System.IO.Stream stream, int maxSizeOfHeaders)
		{
			throw new Exception("The method or operation is not implemented.");
		}

		#endregion

	}

	public class PlainXMLMessage : Message
	{
		private const string MissingRootError = "Root element is missing.";

		private MessageHeaders headers;
		private MessageProperties properties;
		private XmlReader reader;

		public PlainXMLMessage(MessageVersion version, string action, XmlReader body)
			: base()
		{
			headers = new MessageHeaders(version);
			properties = new MessageProperties();
			properties.Add(Constants.HttpRequestMessageProperty, new HttpRequestMessageProperty());
			reader = body;
		}

		public PlainXMLMessage(MessageVersion version, string action)
			: base()
		{
			headers = new MessageHeaders(version);
			properties = new MessageProperties();
		}

		public override MessageHeaders Headers
		{
			get
			{
				return headers;
			}
		}

		protected override void OnWriteBodyContents(XmlDictionaryWriter writer)
		{
			if (reader != null)
			{
				while (!reader.EOF)
				{
					try
					{
						writer.WriteNode(reader, true);
					}
					catch (XmlException exception)
					{
						if (exception.Message.Contains(PlainXMLMessage.MissingRootError))
						{
							return;
						}
						else { throw exception; }
					}
				}

			}
		}

		public override MessageProperties Properties
		{
			get
			{
				return properties;
			}
		}

		public override MessageVersion Version
		{
			get
			{
				return Headers.MessageVersion;
			}
		}
	}

}
