﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Text;
using System.IO;
using System.IO.Compression;
using System.Runtime.Serialization;
using System.Runtime.InteropServices;
using System.Security;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Design;
using System.ServiceModel.Security.Protocols;
using System.Xml;


namespace Microsoft.Samples.Indigo.GZipEncoder
{
    public class CompressionMessageEncoderFactory : MessageEncoderFactory
    {
        #region Private fields
        MessageEncoder encoder;
        #endregion

        #region Constructor
        public CompressionMessageEncoderFactory(MessageEncoderFactory messageEncoderFactory)
        {
            if (messageEncoderFactory == null)
                throw new ArgumentNullException("messageEncoderFactory", "A valid message encoder factory must be passed to the CompressionEncoder");
            encoder = new CompressionMessageEncoder(messageEncoderFactory.Encoder);

        }
        #endregion

        #region Properties
        public override MessageEncoder Encoder
        {
            get { return encoder; }
        }

        public override MessageVersion MessageVersion
        {
            get { return encoder.MessageVersion; }
        }
        #endregion
    }

    class CompressionMessageEncoder :MessageEncoder 
    {
        #region Private fields
        static string compressionContentType = "application/x-gzip";
        MessageEncoder innerEncoder;
        #endregion

        #region Constructor
        internal CompressionMessageEncoder(MessageEncoder messageEncoder)
            : base()
        {
            if (messageEncoder == null)
                throw new ArgumentNullException("messageEncoder", "A valid message encoder must be passed to the CompressionEncoder");
            innerEncoder = messageEncoder;
        }
        #endregion

        #region Properties
        //public override string CharSet
        //{
        //    get { return ""; }
        //}

        public override string ContentType
        {
            get { return compressionContentType; }
        }

        public override string MediaType
        {
            get { return compressionContentType; }
        }

        public override MessageVersion MessageVersion
        {
            get { return innerEncoder.MessageVersion; }
        }
        #endregion

        #region Buffer compression and decompression
        static ArraySegment<byte> CompressBuffer(ArraySegment<byte> buffer, BufferManager bufferManager, int messageOffset)
        {
            MemoryStream memoryStream = new MemoryStream();
            memoryStream.Write(buffer.Array, 0, messageOffset);

            using (GZipStream gzStream = new GZipStream(memoryStream, CompressionMode.Compress, true))
            {
                gzStream.Write(buffer.Array, messageOffset, buffer.Count);
            }


            byte[] compressedBytes = memoryStream.ToArray();
            byte[] bufferedBytes = bufferManager.TakeBuffer(compressedBytes.Length);

            Array.Copy(compressedBytes, 0, bufferedBytes, 0, compressedBytes.Length);

            bufferManager.ReturnBuffer(buffer.Array);
            ArraySegment<byte> byteArray = new ArraySegment<byte>(bufferedBytes, messageOffset, bufferedBytes.Length - messageOffset);

            return byteArray;
        }

        static ArraySegment<byte> DecompressBuffer(ArraySegment<byte> buffer, BufferManager bufferManager)
        {
            MemoryStream memoryStream = new MemoryStream(buffer.Array, buffer.Offset, buffer.Count - buffer.Offset);
            MemoryStream decompressedStream = new MemoryStream();
            int totalRead = 0;
            int blockSize = 1024;
            byte[] tempBuffer = bufferManager.TakeBuffer(blockSize);
            using (GZipStream gzStream = new GZipStream(memoryStream, CompressionMode.Decompress))
            {
                while (true)
                {
                    int bytesRead = gzStream.Read(tempBuffer, 0, blockSize);
                    if (bytesRead == 0)
                        break;
                    decompressedStream.Write(tempBuffer, 0, bytesRead);
                    totalRead += bytesRead;
                }
            }
            bufferManager.ReturnBuffer(tempBuffer);

            byte[] decompressedBytes = decompressedStream.ToArray();
            byte[] bufferManagerBuffer = bufferManager.TakeBuffer(decompressedBytes.Length + buffer.Offset);
            Array.Copy(buffer.Array, 0, bufferManagerBuffer, 0, buffer.Offset);
            Array.Copy(decompressedBytes, 0, bufferManagerBuffer, buffer.Offset, decompressedBytes.Length);

            ArraySegment<byte> byteArray = new ArraySegment<byte>(bufferManagerBuffer, buffer.Offset, decompressedBytes.Length);
            bufferManager.ReturnBuffer(buffer.Array);

            return byteArray;
        }
        #endregion

        #region Buffered message handling
        public override Message ReadMessage(ArraySegment<byte> buffer, BufferManager bufferManager)
        {
            ArraySegment<byte> decompressedBuffer = DecompressBuffer(buffer, bufferManager);
            Message returnMessage = innerEncoder.ReadMessage(decompressedBuffer, bufferManager);
            returnMessage.Properties.Encoder = this;
            return returnMessage;
        }

        public override ArraySegment<byte> WriteMessage(Message message, int maxMessageSize, BufferManager bufferManager, int messageOffset)
        {
            ArraySegment<byte> buffer = innerEncoder.WriteMessage(message, maxMessageSize, bufferManager, messageOffset);
            return CompressBuffer(buffer, bufferManager, messageOffset);
        }
        #endregion

        #region Streaming message handling
        public override Message ReadMessage(System.IO.Stream stream, int maxSizeOfHeaders)
        {
            throw new NotSupportedException("Compression Encoding is not yet supported for streamed communications");
        }


        public override void WriteMessage(Message message, System.IO.Stream stream)
        {
            throw new NotSupportedException("Compression Encoding is not yet supported for streamed communications");
        }
        #endregion

    }

    public sealed class CompressionMessageEncodingBindingElement : BindingElement, IMessageEncodingBindingElement
    {
        #region Private fields
        IMessageEncodingBindingElement innerBindingElement;
        #endregion

        #region Constructor
        public CompressionMessageEncodingBindingElement()
            : this(new TextMessageEncodingBindingElement()) { }

        public CompressionMessageEncodingBindingElement(IMessageEncodingBindingElement messageEncoderBindingElement)
        {
            this.innerBindingElement = messageEncoderBindingElement;
        }
        #endregion

        #region Properties
        public IMessageEncodingBindingElement InnerMessageEncodingBindingElement
        {
            get { return innerBindingElement; }
            set { innerBindingElement = value; }
        }
        #endregion

        #region IMessageEncodingBindingElement methods
        public MessageEncoderFactory CreateMessageEncoderFactory()
        {
            return new CompressionMessageEncoderFactory(innerBindingElement.CreateMessageEncoderFactory());
        }

        public AddressingVersion AddressingVersion
        {
            get { return innerBindingElement.AddressingVersion; }
        }
        #endregion

        #region BindingElement overrides
        public override BindingElement Clone()
        {
            //return ((BindingElement)innerBindingElement);
            return new CompressionMessageEncodingBindingElement(this.innerBindingElement);
        }
        public override ChannelProtectionRequirements GetProtectionRequirements()
        {
            //compression encoder doesn't have any requirements
            //return requirements from inner binding element
            return ((BindingElement)innerBindingElement).GetProtectionRequirements();
        }

        public override IChannelFactory BuildChannelFactory(ChannelBuildContext context)
        {
            if (context == null)
                throw new ArgumentNullException("context");

            context.UnhandledBindingElements.Add(this);
            return context.BuildInnerChannelFactory();
        }

        public override IChannelListener<TChannel> BuildChannelListener<TChannel>(ChannelBuildContext context)
        {
            if (context == null)
                throw new ArgumentNullException("context");

            context.UnhandledBindingElements.Add(this);
            return context.BuildInnerChannelListener<TChannel>();
        }

        public override bool CanBuildChannelListener<TChannel>(ChannelBuildContext context)
        {
            if (context == null)
                throw new ArgumentNullException("context");

            context.UnhandledBindingElements.Add(this);
            return context.CanBuildInnerChannelListener<TChannel>();
        }
        #endregion


    }


    public class CompressionMessageEncodingSection : BindingElementExtensionSection
    {
        #region Constructor
        public CompressionMessageEncodingSection()
        {
        }
        #endregion

        #region Properties
        public override Type BindingElementType
        {
            get { return typeof(CompressionMessageEncodingBindingElement); }
        }
                 

        public override string ConfiguredSectionName
        {
            get { return "compressionMessageEncoding"; }
        }

        [ConfigurationProperty("innerMessageEncoding", DefaultValue = "textMessageEncoding")]
        public string InnerMessageEncoding
        {
            get { return (string)base["innerMessageEncoding"]; }
            set { base["innerMessageEncoding"] = value; }
        }
        #endregion

        #region BindingElementExtensionSection overrides
        public override void ApplyConfiguration(BindingElement bindingElement)
        {
            CompressionMessageEncodingBindingElement binding = (CompressionMessageEncodingBindingElement)bindingElement;
            PropertyInformationCollection propertyInfo = this.ElementInformation.Properties;
            if (propertyInfo["innerMessageEncoding"].ValueOrigin != PropertyValueOrigin.Default)
            {
                switch (this.InnerMessageEncoding)
                {
                    case "textMessageEncoding":
                        binding.InnerMessageEncodingBindingElement = new TextMessageEncodingBindingElement();
                        break;
                    case "binaryMessageEncoding":
                        binding.InnerMessageEncodingBindingElement = new BinaryMessageEncodingBindingElement();
                        break;
                }
            }
        }

        protected override BindingElement CreateBindingElement()
        {
            CompressionMessageEncodingBindingElement bindingElement = new CompressionMessageEncodingBindingElement();
            this.ApplyConfiguration(bindingElement);
            return bindingElement;
        }
        #endregion
    }
}
