// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

#region using
using System;
using System.Collections.Generic;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Design;
using System.Globalization;
#endregion

namespace Microsoft.ServiceModel.Samples
{
    /// <summary>
    /// IChannelFactory implementation for Udp.
    /// 
    /// Supports IOutputChannel only, as Udp is fundamentally
    /// a datagram protocol.
    /// </summary>
    class MailChannelFactory : ChannelFactoryBase
    {
        #region member_variables
        BufferManager bufferManager;
        int maxMessageSize;
        MessageEncoderFactory messageEncoderFactory;
        bool multicast;
        #endregion

        internal MailChannelFactory(MailTransportBindingElement bindingElement, ChannelBuildContext context)
            : base(context.Binding)
        {
            #region populate_members_from_binding_element
            this.fromAddress = bindingElement.FromAddress;

            this.maxMessageSize = bindingElement.MaxMessageSize;
            this.multicast = bindingElement.Multicast;
            this.bufferManager = BufferManager.CreateBufferManager(bindingElement.MaxBufferPoolSize, bindingElement.MaxMessageSize);

            IMessageEncodingBindingElement messageEncoderBindingElement = context.UnhandledBindingElements.Remove<IMessageEncodingBindingElement>();
            if (messageEncoderBindingElement != null)
            {
                this.messageEncoderFactory = messageEncoderBindingElement.CreateMessageEncoderFactory();
            }
            else
            {
                this.messageEncoderFactory = MailConstants.DefaultMessageEncoderFactory;
            }
            #endregion
        }

        #region simple_property_accessors
        string fromAddress = null;

        public string FromAddress
        {
            get
            {
                return this.fromAddress;
            }

            set
            {
                this.fromAddress = value;
            }
        }

        public BufferManager BufferManager
        {
            get
            {
                return this.bufferManager;
            }
        }

        public int MaxMessageSize
        {
            get
            {
                return this.maxMessageSize;
            }
        }

        public MessageEncoderFactory MessageEncoderFactory
        {
            get
            {
                return this.messageEncoderFactory;
            }
        }

        public bool Multicast
        {
            get
            {
                return this.multicast;
            }
        }
        #endregion

        #region IChannelFactory_Methods
        /// <summary>
        /// We only support the default Message Version
        /// </summary>
        public override MessageVersion MessageVersion
        {
            get
            {
                return messageEncoderFactory.MessageVersion;
            }
        }

        /// <summary>
        /// URI scheme associated with the UDP transport 
        /// </summary>
        public override string Scheme
        {
            get
            {
                return MailConstants.Scheme;
            }
        }

        /// <summary>
        /// Used by higher layers to determine what types of channels this
        /// factory supports. Which in this case is just IOutputChannel.
        /// </summary>
        public override bool CanCreateChannel<TChannel>()
        {
            return (typeof(TChannel) == typeof(IOutputChannel));
        }

        /// <summary>
        /// Create a new Udp Channel. Supports IOutputChannel.
        /// </summary>
        /// <typeparam name="TChannel">The type of Channel to create (e.g. IOutputChannel)</typeparam>
        /// <param name="remoteAddress">The address of the remote endpoint</param>
        /// <returns></returns>
        protected override TChannel OnCreateChannel<TChannel>(EndpointAddress remoteAddress, Uri via)
        {
            if (typeof(TChannel) == typeof(IOutputChannel))
            {
                return (TChannel)(object)new MailOutputChannel(this, remoteAddress, via, MessageEncoderFactory.Encoder);
            }

            throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                "Unsupported channel type: {0}.", typeof(TChannel).Name), "TChannel");
        }

        protected override void OnClosed()
        {
            this.bufferManager.Clear();
        }
        #endregion
    }
}
