// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

namespace Microsoft.ServiceModel.Samples
{
    #region using
    using System;
    using System.IO;
    using System.Xml;
    using System.Diagnostics;
    using System.Net;
    using System.Net.Sockets;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Globalization;
    using System.ServiceModel;
    using System.ServiceModel.Channels;
    using System.ServiceModel.Design;
    using System.Text;
    using System.Text.RegularExpressions;
    using System.Threading;
    

    #endregion

    /// <summary>
    /// IListenerFactory implementation for Udp.
    /// 
    /// Supports IInputChannel only, as Udp is fundamentally
    /// a datagram protocol. Uses a listening UDP socket for each
    /// Factory instance, then demuxes to the appropriate IInputListener
    /// based on the registered filters. More advanced implementations could factor
    /// out the listening socket and demux from this listening object
    /// to separate factories based on a framing Via.
    /// </summary>
    class MailChannelListener : ChannelListenerBase<IInputChannel>
    {
        #region member_variables
        BufferManager bufferManager;

        //The UDP network sockets.
        List<Socket> listenSockets;

        int maxMessageSize;
        MessageEncoderFactory messageEncoderFactory;
        bool multicast;

        AsyncCallback onReceive;
        Uri uri;

        InputQueue<IInputChannel> channelQueue;

        //The channel associated with this listener.
        MailInputChannel currentChannel;

        object currentChannelLock;

        string userName = null;
        string password = null;

        #endregion

        internal MailChannelListener(MailTransportBindingElement bindingElement, ChannelBuildContext context)
            : base(context.Binding)
        {
            #region populate_members_from_binding_element

            this.userName = bindingElement.UserName;
            this.password = bindingElement.Password;

            this.maxMessageSize = bindingElement.MaxMessageSize;
            this.multicast = bindingElement.Multicast;
            this.bufferManager = BufferManager.CreateBufferManager(bindingElement.MaxBufferPoolSize, bindingElement.MaxMessageSize);
            IMessageEncodingBindingElement messageEncoderBindingElement = context.UnhandledBindingElements.Remove<IMessageEncodingBindingElement>();
            if (messageEncoderBindingElement != null)
            {
                this.messageEncoderFactory = messageEncoderBindingElement.CreateMessageEncoderFactory();
            }
            else
            {
                this.messageEncoderFactory = MailConstants.DefaultMessageEncoderFactory;
            }
            this.channelQueue = new InputQueue<IInputChannel>();
            this.channelQueue.Open();
            this.currentChannelLock = new object();
            this.listenSockets = new List<Socket>(2);
            #endregion
        }

        #region simple_property_accessors
        internal TimeSpan InternalReceiveTimeout
        {
            get { return this.DefaultReceiveTimeout; }
        }

        public MessageEncoderFactory MessageEncoderFactory
        {
            get
            {
                return messageEncoderFactory;
            }
        }
        #endregion

        #region IListenerFactory_Methods
        public override MessageVersion MessageVersion
        {
            get
            {
                return messageEncoderFactory.MessageVersion;
            }
        }

        public override string Scheme
        {
            get
            {
                return MailConstants.Scheme;
            }
        }

        public override Uri Uri
        {
            get
            {
                return this.uri;
            }
        }
        #endregion

        #region Lifecycle State Machine
        /// <summary>
        /// Shutdown ungracefully
        /// </summary>
        protected override void OnAbort()
        {
            try
            {
                // Abort can be called at anytime, so we can't assume that
                // we've been Opened successfully (and thus may not have any listen sockets)
                lock (this.ThisLock)
                {
                    CloseListenSockets(TimeSpan.Zero);
                }
            }
            finally
            {
                base.OnAbort();
            }
        }

        /// <summary>
        /// Shutdown gracefully
        /// </summary>
        protected override void OnClose(TimeSpan timeout)
        {
            lock (this.ThisLock)
            {
                CloseListenSockets(TimeSpan.Zero);
            }
            base.OnClose(timeout);
        }

        void CloseListenSockets(TimeSpan timeout)
        {
            // TODO: telescoping timeout
            for (int i = 0; i < listenSockets.Count; i++)
            {
                this.listenSockets[i].Close((int)timeout.TotalMilliseconds);
            }
            listenSockets.Clear();
        }

        protected override void OnClosed()
        {
            if (this.bufferManager != null)
            {
                this.bufferManager.Clear();
            }
        }

        /// <summary>
        /// Initialize any objects we're going to need for the opened factory
        /// </summary>
        protected override void OnOpening()
        {
            this.onReceive = new AsyncCallback(this.OnReceive);
        }

        /// <summary>
        /// Open the listener factory for use. Ensures our UDP socket is bound
        /// </summary>
        protected override void OnOpen(TimeSpan timeout)
        {
            if (uri == null)
            {
                throw new InvalidOperationException("Uri must be set before ChannelListener is opened.");
            }

            base.OnOpen(timeout);
            if (this.listenSockets.Count == 0)
            {
                /*
                if (uri.HostNameType == UriHostNameType.IPv6 ||
                    uri.HostNameType == UriHostNameType.IPv4)
                {
                    listenSockets.Add(CreateListenSocket(IPAddress.Parse(uri.Host), uri.Port));
                }
                else
                {
                    listenSockets.Add(CreateListenSocket(IPAddress.Any, uri.Port));
                    if (Socket.OSSupportsIPv6)
                    {
                        listenSockets.Add(CreateListenSocket(IPAddress.IPv6Any, uri.Port));
                    }
                }
                */
            }
        }

        /// <summary>
        /// Open has completed, start an asynchronous receive on our socket.
        /// </summary>
        protected override void OnOpened()
        {
            base.OnOpened();
            Socket[] socketsSnapshot = listenSockets.ToArray();
            WaitCallback startReceivingCallback = new WaitCallback(StartReceiving);
            ThreadPool.QueueUserWorkItem(startReceivingCallback);
            
           
        }
        #endregion


        Socket CreateListenSocket(IPAddress address, int port)
        {
            Socket listenSocket = new Socket(address.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
            listenSocket.Connect(this.uri.Host, this.uri.Port);
            this.listenSockets.Add(listenSocket);
            if (this.ErrorResponse(this.Transmit(null, null)))
            {
                throw new Exception("Failure to connect.");
            }
            return listenSocket;
        }

        EndPoint CreateDummyEndPoint(Socket socket)
        {
            if (socket.AddressFamily == AddressFamily.InterNetwork)
            {
                return new IPEndPoint(IPAddress.Any, 0);
            }
            else
            {
                return new IPEndPoint(IPAddress.IPv6Any, 0);
            }
        }

        /*
        void StartReceiving(object state)
        {
            Socket listenSocket = (Socket)state;
            IAsyncResult result = null;

            try
            {
                lock (ThisLock)
                {
                    if (base.State == CommunicationState.Opened)
                    {
                        EndPoint dummy = CreateDummyEndPoint(listenSocket);
                        byte[] buffer = this.bufferManager.TakeBuffer(maxMessageSize);
                        result = listenSocket.BeginReceiveFrom(buffer, 0, buffer.Length,
                            SocketFlags.None, ref dummy, this.onReceive, new SocketReceiveState(listenSocket, buffer));
                    }
                }

                if (result != null && result.CompletedSynchronously)
                {
                    ContinueReceiving(result, listenSocket);
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error in receiving from the socket.");
                Debug.WriteLine(e.ToString());
            }
        }
        */

        private void StartReceiving(object state)
        {
            try
            {
                while (base.State == CommunicationState.Opened)
                {
                    lock (ThisLock)
                    {
                        if (this.listenSockets.Count == 0)
                        {
                            IPHostEntry host = Dns.GetHostEntry(uri.Host);
                            IPAddress address = host.AddressList[0];
                            this.CreateListenSocket(address, this.uri.Port);
                        }

                        Console.WriteLine("Doing reception.");
                        this.POP3Reception();


                    }
                    Thread.Sleep(5000);
                }
            }
            catch (Exception exception)
            {
                Console.WriteLine("Error in receiving from the socket.");
                Console.WriteLine(exception.ToString());
            }
        }

        private void POP3Reception()
        {
            const string LineTerminator = "\r\n";
            const string MessageTerminator = ".";

            StringBuilder messageBuffer = null;
            Regex regularExpression = new Regex(@"(^\+OK\s)(\d*)([\s])(\d*)");
            Match match = null;
            int messageCount;
            int mailDropSize;
            string mailMessage = null;
            Message message = null;
            byte[] buffer = null;
            this.LogIn();

            string response = null;
            response = this.Transmit("STAT", null);
            if (this.ErrorResponse(response))
            {
                throw new Exception("Error retrieving count of new messages.");
            }
            if (!(regularExpression.IsMatch(response)))
            {
                throw new Exception("Error parsing count of new messages.");
            }
            match = regularExpression.Match(response);
            messageCount = int.Parse(match.Groups[2].Value);
            mailDropSize = int.Parse(match.Groups[4].Value);
            Console.WriteLine("Message count: {0}", messageCount);

            for (int messageIndex = 1; 
                messageIndex <= messageCount; 
                messageIndex++)
            {
                messageBuffer = new StringBuilder();
                while (true)
                {
                    response = this.Transmit(
                        string.Format("RETR {0}", messageIndex), 
                        mailDropSize);
                    if (this.ErrorResponse(response))
                    {
                        throw new Exception("Error retrieving message.");
                    }

                    if (response.StartsWith(MessageTerminator))
                    {
                        if (!(response.EndsWith(
                            MessageTerminator + LineTerminator)))
                        {
                            response = response.Substring(1);
                        }
                    }
                    if (response.EndsWith(MessageTerminator + LineTerminator))
                    {
                        messageBuffer.Append(
                            response.Substring(
                                0, 
                                response.Length - LineTerminator.Length));
                        break;
                    }

                    messageBuffer.Append(response);
                }
                mailMessage = messageBuffer.ToString();

                response = this.Transmit(
                    string.Format("DELE {0}", messageIndex), 
                    null);
                if (this.ErrorResponse(response))
                {
                    throw new Exception("Error deleting message.");
                }

                buffer = Encoding.ASCII.GetBytes(mailMessage);
                message = messageEncoderFactory.Encoder.ReadMessage(
                    new ArraySegment<byte>(buffer, 0, buffer.Length), 
                    bufferManager);
                if (message != null)
                {
                    ThreadPool.QueueUserWorkItem(
                        new WaitCallback(DispatchCallback), message);
                }
                else
                {
                    EventLog.WriteEntry(
                        MailConstants.EventLogSourceName, 
"A message was dropped because it was not in the expected format.", 
                        EventLogEntryType.Warning);
                }
            }

            response = this.Transmit("QUIT", null);
            if (this.ErrorResponse(response))
            {
                throw new Exception("Error updating mailbox.");
            }

            this.CloseListenSockets(TimeSpan.Zero);
        }

        private bool ErrorResponse(string response)
        {
            if (response.StartsWith("-ERR"))
            {
                return true;
            }
            return false;
        }

        private void LogIn()
        {
            if (this.ErrorResponse(
                this.Transmit(string.Format("USER {0}", this.userName), null)))
            {
                throw new Exception("Username rejected.");
            }
            if (this.ErrorResponse(
                this.Transmit(string.Format("PASS {0}", this.password), null)))
            {
                throw new Exception("Password rejected.");
            }
        }

        private string Transmit(string message, int? bufferSize)
        {
            if (message != null)
            {
                this.listenSockets[0].Send(
                    Encoding.ASCII.GetBytes(string.Concat(message, "\r\n")));
            }

            byte[] buffer = new byte[
                bufferSize != null ? bufferSize.Value + 512 : 512];
            int bytesRead = this.listenSockets[0].Receive(buffer);
            StringBuilder builder = new StringBuilder();
            builder.Append(Encoding.ASCII.GetChars(buffer), 0, bytesRead);
            return builder.ToString();
        }

        void ContinueReceiving(IAsyncResult receiveResult, Socket listenSocket)
        {
            bool continueReceiving = true;

            while (continueReceiving)
            {
                Message receivedMessage = null;

                if (receiveResult != null)
                {
                    receivedMessage = EndReceive(listenSocket, receiveResult);
                    receiveResult = null;
                }

                lock (ThisLock)
                {
                    if (base.State == CommunicationState.Opened)
                    {
                        EndPoint dummy = CreateDummyEndPoint(listenSocket);
                        byte[] buffer = this.bufferManager.TakeBuffer(maxMessageSize);
                        receiveResult = listenSocket.BeginReceiveFrom(buffer, 0, buffer.Length,
                            SocketFlags.None, ref dummy, this.onReceive, new SocketReceiveState(listenSocket, buffer));
                    }
                }

                if (receiveResult == null || !receiveResult.CompletedSynchronously)
                {
                    continueReceiving = false;
                    Dispatch(receivedMessage);
                }
                else if (receivedMessage != null)
                {
                    ThreadPool.QueueUserWorkItem(new WaitCallback(DispatchCallback), receivedMessage);
                }
            }
        }

        Message EndReceive(Socket listenSocket, IAsyncResult result)
        {
            // if we've started the shutdown process, then we've disposed
            // the socket and calls to socket.EndReceive will throw 
            if (base.State != CommunicationState.Opened)
                return null;

            byte[] buffer = ((SocketReceiveState)result.AsyncState).Buffer;
            Debug.Assert(buffer != null);
            Message message = null;

            try
            {
                int count = 0;

                lock (ThisLock)
                {
                    // if we've started the shutdown process, socket is disposed
                    // and calls to socket.EndReceive will throw 
                    if (base.State == CommunicationState.Opened)
                    {
                        EndPoint dummy = CreateDummyEndPoint(listenSocket);
                        count = listenSocket.EndReceiveFrom(result, ref dummy);
                    }
                }

                if (count > 0)
                {
                    message = MessageEncoderFactory.Encoder.ReadMessage(new ArraySegment<byte>(buffer, 0, count), bufferManager);
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error in completing the async receive via EndReceiveFrom method.");
                Debug.WriteLine(e.ToString());
            }
            finally
            {
                if (message == null)
                {
                    this.bufferManager.ReturnBuffer(buffer);
                    buffer = null;
                }
            }

            return message;
        }

        //Called when an ansynchronous receieve operation completes
        //on the listening socket.
        void OnReceive(IAsyncResult result)
        {
            if (result.CompletedSynchronously)
                return;

            ContinueReceiving(result, ((SocketReceiveState)result.AsyncState).Socket);
        }

        void DispatchCallback(object state)
        {
            Dispatch((Message)state);
        }

        /// <summary>
        /// Matches an incoming message to its waiting listener,
        /// using the FilterTable to dispatch the message to the correc
        /// listener. If no listener is waiting for the message, it is silently
        /// discarded.
        /// </summary>
        void Dispatch(Message message)
        {
            if (message == null)
                return;

            try
            {
                MailInputChannel newChannel;
                bool channelCreated = CreateOrRetrieveChannel(out newChannel);

                newChannel.Dispatch(message);

                if (channelCreated)
                {
                    //Hand the channel off to whomever is waiting for AcceptChannel()
                    //to complete
                    this.channelQueue.EnqueueAndDispatch(newChannel);
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error dispatching Message.");
                Debug.WriteLine(e.ToString());
            }
        }

        /// <summary>
        /// Used to get a unique uri (by CompositeDuplexChannelFactory for example).
        /// We get a unique TCP port by binding to "port 0"
        /// </summary>
        public override void SetUniqueUri(string host)
        {
            if (host == null)
            {
                throw new ArgumentNullException("host");
            }

            int port;

            lock (this.ThisLock)
            {
                CloseListenSockets(TimeSpan.Zero);
                IPAddress ipAddress = null;
                if (IPAddress.TryParse(host, out ipAddress))
                {
                    Socket socket = CreateListenSocket(ipAddress, 0);
                    port = ((IPEndPoint)socket.LocalEndPoint).Port;
                    listenSockets.Add(socket);
                }
                else
                {
                    Socket socket = CreateListenSocket(IPAddress.Any, 0);
                    port = ((IPEndPoint)socket.LocalEndPoint).Port;
                    listenSockets.Add(socket);
                    if (Socket.OSSupportsIPv6)
                    {
                        listenSockets.Add(CreateListenSocket(IPAddress.IPv6Any, port));
                    }
                }
            }

            UriBuilder uriBuilder = new UriBuilder(Scheme, host, port);
            SetUri(uriBuilder.Uri);
        }

        public override void SetUniqueUri(string host, int port)
        {
            throw new NotSupportedException();
        }

        public override void SetUri(Uri baseAddress, string relativeAddress)
        {
            if (baseAddress == null)
                throw new ArgumentNullException("baseAddress");

            if (relativeAddress == null)
                throw new ArgumentNullException("relativeAddress");

            if (!baseAddress.IsAbsoluteUri)
                throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                    "Base address must be an absolute URI."), "baseAddress");

            if (baseAddress.Scheme != this.Scheme)
            {
                // URI schemes are case-insensitive, so try a case insensitive compare now
                if (string.Compare(baseAddress.Scheme, this.Scheme, true, System.Globalization.CultureInfo.InvariantCulture) != 0)
                {
                    throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                        "Invalid URI scheme: {0}.", baseAddress.Scheme), "baseAddress");
                }
            }

            Uri fullUri = baseAddress;

            // Ensure that baseAddress Path does end with a slash if we have a relative address
            if (relativeAddress != string.Empty)
            {
                if (!baseAddress.AbsolutePath.EndsWith("/"))
                {
                    UriBuilder uriBuilder = new UriBuilder(baseAddress);
                    uriBuilder.Path = uriBuilder.Path + "/";
                    baseAddress = uriBuilder.Uri;
                }

                fullUri = new Uri(baseAddress, relativeAddress);
            }

            lock (base.ThisLock)
            {
                ThrowIfDisposedOrImmutable();
                this.uri = fullUri;
                CloseListenSockets(TimeSpan.Zero);
            }
        }

        //Synchronously returns a channel that is attached to this listener.
        public override IInputChannel AcceptChannel(TimeSpan timeout)
        {
            if (!this.IsDisposed)
            {
                this.EnsureChannelAvailable();
            }

            IInputChannel channel;
            if (this.channelQueue.TryDequeue(timeout, out channel))
            {
                return channel;
            }
            else
            {
                throw new TimeoutException();
            }
        }

        public override IAsyncResult BeginAcceptChannel(TimeSpan timeout, AsyncCallback callback, object state)
        {
            if (!this.IsDisposed)
            {
                this.EnsureChannelAvailable();
            }

            return this.channelQueue.BeginTryDequeue(timeout, callback, state);
        }

        public override IInputChannel EndAcceptChannel(IAsyncResult result)
        {
            IInputChannel channel;
            if (this.channelQueue.EndTryDequeue(result, out channel))
            {
                return channel;
            }
            else
            {
                throw new TimeoutException();
            }
        }

        public override bool WaitForChannel(TimeSpan timeout)
        {
            throw new NotImplementedException();
        }

        public override IAsyncResult BeginWaitForChannel(TimeSpan timeout, AsyncCallback callback, object state)
        {
            throw new NotImplementedException();
        }

        public override bool EndWaitForChannel(IAsyncResult result)
        {
            throw new NotImplementedException();
        }

        //Guarantees that channel is attached to this listener.
        void EnsureChannelAvailable()
        {
            MailInputChannel newChannel;
            bool channelCreated = CreateOrRetrieveChannel(out newChannel);

            if (channelCreated)
            {
                this.channelQueue.EnqueueAndDispatch(newChannel);
            }
        }

        bool CreateOrRetrieveChannel(out MailInputChannel newChannel)
        {
            bool channelCreated = false;

            if ((newChannel = currentChannel) == null)
            {
                lock (currentChannelLock)
                {
                    if ((newChannel = currentChannel) == null)
                    {
                        newChannel = new MailInputChannel(this);
                        newChannel.Closed += new EventHandler(this.OnChannelClosed);
                        currentChannel = newChannel;
                        channelCreated = true;
                    }
                }
            }

            return channelCreated;
        }

        void OnChannelClosed(object sender, EventArgs args)
        {
            MailInputChannel channel = (MailInputChannel)sender;

            lock (this.currentChannelLock)
            {
                if (channel == this.currentChannel)
                {
                    this.currentChannel = null;
                }
            }
        }

        class SocketReceiveState
        {
            Socket socket;
            byte[] buffer;
            public SocketReceiveState(Socket socket, byte[] buffer)
            {
                this.socket = socket;
                this.buffer = buffer;
            }

            public Socket Socket
            {
                get { return this.socket; }
            }

            public byte[] Buffer
            {
                get { return this.buffer; }
            }
        }
    }
}
