// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

#region using
using System;
using System.Globalization;
using System.ServiceModel;
using System.ServiceModel.Design;
using System.ServiceModel.Security.Protocols;
using System.Xml;
using WsdlNS = System.Web.Services.Description;
using System.Collections.Generic;
#endregion

namespace Microsoft.ServiceModel.Samples
{
    /// <summary>
    /// Udp Binding Element.  
    /// Used to configure and construct Udp ChannelFactories and ChannelListeners.
    /// </summary>
    public class MailTransportBindingElement 
        : BindingElement, ITransportBindingElement // to signal that we're a transport
        , IPolicyExporter // for policy export
    {
        #region member_variables
        long maxBufferPoolSize;
        int maxMessageSize;
        bool multicast;
        #endregion

        #region constructors
        public MailTransportBindingElement()
        {
            this.maxBufferPoolSize = MailDefaults.MaxBufferPoolSize;
            this.maxMessageSize = MailDefaults.MaxMessageSize;
            this.multicast = MailDefaults.Multicast;
        }

        protected MailTransportBindingElement(MailTransportBindingElement other)
            : base(other)
        {
            this.maxBufferPoolSize = other.maxBufferPoolSize;
            this.maxMessageSize = other.maxMessageSize;
            this.multicast = other.multicast;

            this.fromAddress = other.fromAddress;
            this.userName = other.userName;
            this.password = other.password;
        }
        #endregion

        #region CONFIGURATION_Properties
        private string userName = null;
        private string password = null;
        private string fromAddress = null;

        public string FromAddress
        {
            get
            {
                return this.fromAddress;
            }

            set
            {
                this.fromAddress = value;
            }
        }

        public string UserName
        {
            get
            {
                return this.userName;
            }

            set
            {
                this.userName = value;
            }

        }

        public string Password
        {
            get
            {
                return this.password;
            }

            set
            {
                this.password = value;
            }
        }


        public long MaxBufferPoolSize
        {
            get
            {
                return this.maxBufferPoolSize;
            }
            set
            {
                if (value < 0)
                {
                    throw new ArgumentOutOfRangeException("value");
                }
                this.maxBufferPoolSize = value;
            }
        }

        public int MaxMessageSize
        {
            get
            {
                return this.maxMessageSize;
            }

            set
            {
                if (value <= 0)
                {
                    throw new ArgumentOutOfRangeException("value");
                }

                this.maxMessageSize = value;
            }
        }

        public bool Multicast
        {
            get
            {
                return this.multicast;
            }

            set
            {
                this.multicast = value;
            }
        }
        #endregion

        #region BUILD_Methods
        public override IChannelFactory BuildChannelFactory(ChannelBuildContext context)
        {
            if (context == null)
            {
                throw new ArgumentNullException("context");
            }

            return new MailChannelFactory(this, context);
        }

        public override IChannelListener<TChannel> BuildChannelListener<TChannel>(ChannelBuildContext context)
        {
            if (context == null)
            {
                throw new ArgumentNullException("context");
            }

            if (!this.CanBuildChannelListener<TChannel>(context))
            {
                throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, "Unsupported channel type: {0}.", typeof(TChannel).Name));
            }

            return (IChannelListener<TChannel>)(object)new MailChannelListener(this, context);
        }

        /// <summary>
        /// Used by higher layers to determine what types of channel factories this
        /// binding element supports. Which in this case is just IOutputChannel.
        /// </summary>
        public override bool CanBuildChannelFactory<TChannel>(ChannelBuildContext context)
        {
            return (typeof(TChannel) == typeof(IOutputChannel));
        }

        /// <summary>
        /// Used by higher layers to determine what types of channel listeners this
        /// binding element supports. Which in this case is just IInputChannel.
        /// </summary>
        public override bool CanBuildChannelListener<TChannel>(ChannelBuildContext context)
        {
            return (typeof(TChannel) == typeof(IInputChannel));
        }
        #endregion

        #region ITransportBindingElement_members
        string ITransportBindingElement.Scheme
        {
            get
            {
                return MailConstants.Scheme;
            }
        }
        #endregion

        #region Policy_Export
        // We expose in policy The fact that we're UDP, and whether we're multicast or not.
        // Import is done through UdpBindingElementConverter.
        void IPolicyExporter.ExportPolicy(MetadataExporter exporter, PolicyConversionContext context)
        {
            if (exporter == null)
            {
                throw new ArgumentNullException("exporter");
            }

            if (context == null)
            {
                throw new ArgumentNullException("context");
            }

            ICollection<XmlElement> bindingAssertions = context.GetBindingAssertions();
            XmlDocument xmlDocument = new XmlDocument();
            bindingAssertions.Add(xmlDocument.CreateElement(
                MailPolicyStrings.Prefix, MailPolicyStrings.TransportAssertion, MailPolicyStrings.UdpNamespace));

            if (Multicast)
            {
                bindingAssertions.Add(xmlDocument.CreateElement(
                    MailPolicyStrings.Prefix, MailPolicyStrings.MulticastAssertion, MailPolicyStrings.UdpNamespace));
            }

        }
        #endregion

        #region BindingElement_infrastructure_overrides
        public override BindingElement Clone()
        {
            return new MailTransportBindingElement(this);
        }

        public override ChannelProtectionRequirements GetProtectionRequirements()
        {
            return null;
        }
        #endregion
    }
}
