using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Cryptography;
using System.Security.Cryptography.Xml;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Xml;

namespace RepresentationalStateTransfer
{
	public enum ContentRating
	{
		Unknown = -1,
		AllAges = 0,
		AdultsOnly = 1
	}

	public class ReallySimpleRatedSyndication
	{
		private const string personalCertificateStoreLocation = @"My";
		private const string trustedPeopleCertificateStoreLocation = @"TrustedPeople";
		private const string contentRatingCertificateSubjectName = @"Woodgrove";
		private const string contentRatingExtensionTagName = @"contentRating";
		private const string linkTagName = @"link";
		private const string signedContentRatingExtensionTagName = @"signedContentRating";
		private const string contentRatingExtensionNamespace = @"http://blog.code.com/ContentRating";
		private const string signatureValueTagName = "SignatureValue";
		private const string contentRatingIdentifier = "ContentRatingIdentifier";
		private const string linkIdentifier = "LinkIdentifier";
		private const string universalResourceIdentifierPrefix = @"#";
		private const string signatureTagName = @"Signature";
		private const string xmlDigitalSignatureNamespace = @"http://www.w3.org/2000/09/xmldsig#";

		private static XmlElement CreateWrappedSignature(AsymmetricAlgorithm signingKey, XmlElement contentToSign, string identifier)
		{
			SignedXml signedXML = new SignedXml(contentToSign);
			signedXML.SigningKey = signingKey;
			DataObject dataObject = new DataObject(identifier, null, null, contentToSign);
			signedXML.AddObject(dataObject);
			Reference reference = new Reference();
			reference.Uri = string.Concat(ReallySimpleRatedSyndication.universalResourceIdentifierPrefix,identifier);
			signedXML.AddReference(reference);
			signedXML.ComputeSignature();
			return signedXML.GetXml();
		}

		private static bool VerifySignature(X509Certificate2 certificate, XmlElement signatureElement)
		{
			SignedXml signedXML = new SignedXml(signatureElement);
			signedXML.LoadXml(signatureElement);

			bool checks = signedXML.CheckSignature(certificate, true);

			return checks;
		}

		public static void AddSignedContentRating(XmlDocument document)
		{
			X509Store store = new X509Store(ReallySimpleRatedSyndication.personalCertificateStoreLocation, StoreLocation.LocalMachine);
			store.Open(OpenFlags.ReadOnly);
			X509CertificateCollection certificateCollection = store.Certificates.Find(X509FindType.FindBySubjectName, ReallySimpleRatedSyndication.contentRatingCertificateSubjectName, false);
			X509Certificate2 certificate = (X509Certificate2)certificateCollection[0];

			XmlElement ratingElement = (XmlElement)document.GetElementsByTagName(ReallySimpleRatedSyndication.contentRatingExtensionTagName, ReallySimpleRatedSyndication.contentRatingExtensionNamespace).Item(0);
			XmlElement signedElement = ReallySimpleRatedSyndication.CreateWrappedSignature(certificate.PrivateKey, ratingElement, ReallySimpleRatedSyndication.contentRatingIdentifier);
			ratingElement.InnerXml = signedElement.OuterXml;

			XmlElement linkElement = (XmlElement)document.GetElementsByTagName(ReallySimpleRatedSyndication.linkTagName).Item(0);
			signedElement = ReallySimpleRatedSyndication.CreateWrappedSignature(certificate.PrivateKey, linkElement, ReallySimpleRatedSyndication.linkIdentifier);
			linkElement.InnerXml = signedElement.OuterXml;
			
		}

		public static bool ValidateSignedContentRating(XmlDocument document, out ContentRating rating, out string link)
		{
			rating = ContentRating.Unknown;
			link = string.Empty;

			X509Store store = new X509Store(ReallySimpleRatedSyndication.trustedPeopleCertificateStoreLocation, StoreLocation.CurrentUser);
			store.Open(OpenFlags.ReadOnly);
			X509CertificateCollection certificateCollection = store.Certificates.Find(X509FindType.FindBySubjectName, ReallySimpleRatedSyndication.contentRatingCertificateSubjectName, true);
			X509Certificate2 certificate = (X509Certificate2)certificateCollection[0];

			XmlNodeList nodes = document.GetElementsByTagName(ReallySimpleRatedSyndication.signatureTagName, ReallySimpleRatedSyndication.xmlDigitalSignatureNamespace);
			foreach(XmlNode signedNode in nodes)
			{
				if(!(ReallySimpleRatedSyndication.VerifySignature(certificate,(XmlElement)signedNode)))
				{
					return false;
				}
			}
						
			nodes = document.GetElementsByTagName("Object");
			foreach (XmlNode node in nodes)
			{
				foreach (XmlAttribute attribute in node.Attributes)
				{
					switch (attribute.Value)
					{
						case "ContentRatingIdentifier":
							rating = (ContentRating)Enum.Parse(typeof(ContentRating), ((XmlElement)node).InnerText);
							break;
						case "LinkIdentifier":
							link = ((XmlElement)node).InnerText;
							break;
					}
					
				}
			}
				
			return true;
			
		}

	}
}
