using System;
using System.Collections.ObjectModel;
using System.Collections.Generic;
using System.Configuration;
using System.Runtime.Serialization;
using System.Security.Authorization;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.ServiceModel;
using System.ServiceModel.Security;
using System.ServiceModel.Security.Protocols;
using System.ServiceModel.Security.Tokens;
using System.Text;
using System.Threading;
using System.Workflow.Runtime;
using System.Workflow.Runtime.Hosting;
using System.Xml;

namespace SecurityTokenService
{
    public class AuthorizationPolicy : IAuthorizationPolicy
    {
        private string claimMapLocation = null;
        private Dictionary<string, AutoResetEvent>
            waitHandles = new Dictionary<string, AutoResetEvent>();
        private Dictionary<string, string[]> outputClaims =
            new Dictionary<string, string[]>();
        private object waitHandlesLock = new object();
        private object outputClaimsLock = new object();

        public AuthorizationPolicy()
        {
            this.claimMapLocation =
                ConfigurationManager.AppSettings["ClaimMapLocation"];
        }


        public ClaimSet Issuer
        {
            get
            {
                return DefaultClaimSet.System.Issuer;
            }
        }

        public void ClaimMappingCompleted(object sender,
            WorkflowCompletedEventArgs e)
        {
            string requestIdentifier =
                (string)e.OutputParameters[
                "RequestIdentifier"];
            string[] outputClaims = (
                string[])e.OutputParameters[
                "OutputClaims"];
            lock (this.outputClaimsLock)
            {
                this.outputClaims.Add(requestIdentifier, outputClaims);
            }
            AutoResetEvent waitHandle = null;
            lock (this.waitHandlesLock)
            {
                this.waitHandles.TryGetValue(requestIdentifier, out waitHandle);
                if (waitHandle != null)
                {
                    waitHandle.Set();
                }
                this.waitHandles.Remove(requestIdentifier);
            }
        }

        private string[] MapClaims(string[] inputClaims)
        {
            if (Thread.CurrentThread.Name == null)
            {
                Thread.CurrentThread.Name = Guid.NewGuid().ToString();
            }

            using (WorkflowRuntime workflowRuntime = new WorkflowRuntime())
            {
                workflowRuntime.StartRuntime();

                workflowRuntime.WorkflowCompleted += this.ClaimMappingCompleted;

                Type type = typeof(ClaimMappingWorkflow);

                Dictionary<string, object> parameters = 
                    new Dictionary<string, object>();
                parameters.Add(
                    "RequestIdentifier",
                    Thread.CurrentThread.Name);
                parameters.Add(
                    "InputClaims",
                    inputClaims);

                AutoResetEvent waitHandle = new AutoResetEvent(false);

                lock (this.waitHandlesLock)
                {
                    this.waitHandles.Add(Thread.CurrentThread.Name, waitHandle);
                }

                workflowRuntime.CreateWorkflow(type, parameters).Start();

                waitHandle.WaitOne();

                workflowRuntime.StopRuntime();
            }

            string[] outputClaims = null;
            lock (this.outputClaimsLock)
            {
                this.outputClaims.TryGetValue(
                    Thread.CurrentThread.Name,
                    out outputClaims);
                this.outputClaims.Remove(
                    Thread.CurrentThread.Name);
            }

            return outputClaims;
        }


        #region IAuthorizationPolicy Members

        bool IAuthorizationPolicy.Evaluate(EvaluationContext evaluationContext,
            ref object state)
        {

            List<Claim> claimsToAdd = new List<Claim>();

            List<string> inputClaims = new List<string>();

            AuthorizationContext context =
OperationContext.Current.ServiceSecurityContext.AuthorizationContext;
            for (int index = 0; index < context.ClaimSets.Count; index++)
            {
                foreach (Claim claim in context.ClaimSets[index].FindClaims(
"http://schemas.microsoft.com/xsi/2005/05/ClaimType/:Role", null))
                {
                    inputClaims.Add(claim.Resource.ToString());
                }
            }

            string[] roleClaims = this.MapClaims(inputClaims.ToArray());
            Claim targetClaim = null;
            foreach (string roleClaim in roleClaims)
            {
                targetClaim = new Claim(ClaimTypes.Role, roleClaim,
                    Rights.PossessProperty);
                claimsToAdd.Add(targetClaim);
            }

            if (claimsToAdd.Count > 0)
            {
                evaluationContext.AddToTarget(this, new DefaultClaimSet(
                    this.Issuer, claimsToAdd));
            }


            return true;
        }


        ClaimSet IAuthorizationPolicy.Issuer
        {
            get
            {
                return this.Issuer;
            }
        }


        #endregion



        #region IAuthorizationComponent Members

        string IAuthorizationComponent.Id
        {
            get
            {
                return Guid.NewGuid().ToString();
            }
        }

        #endregion
    }


}